@@ -1,9 +1,12 @@
 """Manage containers."""
 import os
 import pystache
+import re
 import subprocess
 import yaml
 from .errors import ConfigError, ContainerError, TestError
+from .tools import Tool
+from .version import Version, get_best_version, get_all_versions, latest as latest_version
 
 config_defaults = {
     'user': None,
@@ -15,6 +18,15 @@ config_defaults = {
 }
 
 
+def first_file(path, filenames):
+    """Returns the first file from `filenames` with read access found in `path`."""
+    for filename in filenames:
+        filepath = os.path.join(path, filename)
+        if os.path.isfile(filepath) and os.access(filepath, os.R_OK):
+            return filepath
+    return None
+
+
 def read_config(configfile):
     """
     Return container configuration from a file. See `Container.from_path` for a description of the
@@ -27,6 +39,8 @@ def read_config(configfile):
         raise ConfigError("failed to open '{}': {}".format(configfile, e))
     except yaml.YAMLError as e:
         raise ConfigError("failed to parse '{}': {}".format(configfile, e))
+    if not isinstance(filedata, dict):
+        raise ConfigError("failed to parse '{}': contains invalid data".format(configfile))
 
     config = config_defaults.copy()
     config.update(filedata)
@@ -39,14 +53,14 @@ def read_config(configfile):
     if '/' in name:
         config['user'], config['name'] = name.split('/', 1)
 
-    for key in ('user', 'name', 'prebuild', 'path'):
+    for key in ('user', 'prebuild', 'path'):
         if config[key] is not None and not isinstance(config[key], (str, unicode)):
-            raise ConfigError("{} must be a string".format(config, key))
+            raise ConfigError("{} must be a string".format(key))
 
     if config['running'] is not None and not isinstance(config['running'], dict):
         raise ConfigError("running must be a dictionary")
     if config['testing'] is not None and not isinstance(config['testing'], dict):
-        raise ConfigError("running must be a dictionary")
+        raise ConfigError("testing must be a dictionary")
 
     return config
 
@@ -55,13 +69,18 @@ class Container(object):
     """Manage a container."""
 
     @classmethod
-    def from_path(cls, path=None, **override):
+    def from_path(cls, path=None, verify=None, **override):
         """
-        Return a container found at the provided directory `path`. Expects a `container.yml` to
-        exist in `path`. If `path` is `None` (the default) the output of `os.getcwd()` will be used
-        for `path`.
+        Return a container found at the provided directory `path`. Expects a container config file
+        to exist in `path`. If `path` is `None` (the default) the output of `os.getcwd()` will be
+        used for `path`. Verification may be enabled by setting `verify` to `True`. Verification
+        causes only verified versions to be built or pushed. Most methods and properties will raise
+        a `ContainerError` if called with verification enabled when not verified version is
+        present.
 
-        The config file must contain valid YAML. Valid keys are:
+        The container config file must be named `container.yml` or `.container.yml`. The first
+        readable file found, in that order, is used. The config file must contain valid YAML. Valid
+        keys are:
 
         * `path` - The path to the directory containing that container's Dockerfile and other
             configuration. This is set to the `path` value determined by this function if absent
@@ -112,14 +131,18 @@ class Container(object):
         if not os.path.isdir(path):
             raise ConfigError("path '{}' is not a directory".format(path))
 
-        configfile = os.path.join(path, 'container.yml')
-        if not os.path.isfile(configfile):
-            raise ConfigError("config file '{}' not found".format(configfile))
+        configfile = first_file(path, ('container.yml', '.container.yml'))
+        if not configfile:
+            raise ConfigError("config file not found")
 
         config = read_config(configfile)
         if config.get('path') is None:
             config['path'] = path
+        if 'version' in config:
+            # manually defined verions are assumed to be verified
+            config['version'] = Version(config['version'], True)
         config.update(override)
+        config['verify'] = verify
         return cls.from_dict(config)
 
     @classmethod
@@ -130,19 +153,23 @@ class Container(object):
             data.get('user'),
             data.get('name'),
             data.get('version'),
+            data.get('verify'),
             data.get('prebuild'),
             data.get('running'),
             data.get('testing'),
         )
 
-    def __init__(self, path, user, name, version=None, prebuild=None, running=None, testing=None):
+    def __init__(self, path, user, name, version=None, verify=None, prebuild=None,
+                 running=None, testing=None):
         """
         Initialize the container. The `path` should be a directory containing the container's
         Dockerfile. The `user` and `name` is combined to create the image repo. The `name` should
         match the directory containing the container's Dockerfile.
 
         The `version` value is used to tag the container on build. If `version` is `None` the
-        version will be detected using `containerctl.version.detect`.
+        version will be detected using `containerctl.version` module. If `verify` is set to `True`
+        then only verified versions will be used to tag the container. If no verified versions are
+        present then a `ContainerError` will be raised when building or pushing.
 
         The `prebuild` param is a command to run before calling `docker build` on the container. If
         `prebuild` is `None` and an executable file called `prebuild` exists in `path` it will be
@@ -170,79 +197,53 @@ class Container(object):
         self.user = user
         self.name = name
 
+        self.verify = verify
         self._version = version
         self._prebuild = prebuild
         self._running = running
         self._testing = testing
 
+        fig_project = (self.user or '') + self.name
+        self._fig_config = os.path.join(self.path, 'fig.yml')
+
+        self._fig = Tool('fig', '-p', fig_project, '-f', self._fig_config, cwd=self.path)
+        self._fig_remove = self._fig('rm', '--force', '-v')
+        self._docker = Tool('docker', cwd=self.path)
+
         # some CI systems do not allow image removal
         self.allow_remove = True
         if os.getenv('CIRCLECI', None):
             self.allow_remove = False
 
-    def to_dict(self, raw=False):
-        """
-        Return the container as a dictionary. If `raw` is `True` then the un-evaluated
-        configuraiton is returned.
-        """
-        def rel(base, path):
-            if path and path.startswith(base):
-                path = '.' + path[len(base):]
-            return path
-
-        cwd = os.getcwd()
-        if raw:
-            data = {'name': self.name}
-            if self.user:
-                data['user'] = self.user
-            if self._version:
-                data['version'] = self._version
-            if self._prebuild:
-                data['prebuild'] = self._prebuild
-            if self._running:
-                data['running'] = self._running
-            if self._testing:
-                data['testing'] = self._testing
-
-            path = rel(cwd, self.path)
-            if path and path != '.':
-                data['path'] = path
-        else:
-            data = {
-                'tag': self.tag,
-                'user': self.user,
-                'name': self.name,
-                'version': self.version,
-                'running': self._get_fig_cfg(False),
-                'testing': self._get_fig_cfg(True),
-                'is-built': self.is_built(),
-                'is-running': self.is_running(),
-                'is-verified': self.is_verified(),
-            }
-            path = rel(cwd, self.path)
-            if path and path != '.':
-                data['path'] = path
-            prebuild = rel(cwd, self._get_prebuild())
-            if prebuild and prebuild != '.':
-                data['prebuild'] = prebuild
-        return data
-
-    def to_config(self):
-        """Return the container's configuration."""
-
     def _format(self, value):
         """
         Format a string value with the `user`, `name`, `version`, and `tag`. Keyword args may be
         provided to override these values or add additonal values for substitution.
+
+        Some simple filters are available for each value. These can be given by appending the value
+        name with '|filter' where filter is the name of the filter. These filters currently exist:
+
+            fig - Convert the value into a name suitable for a fig container.
         """
         if not value:
             return ''
-        return pystache.render(value, {
+
+        patterns = [
+            ('fig', re.compile(r'[^a-zA-Z0-9]'), ''),
+        ]
+
+        context = {
             'user': self.user or '',
             'name': self.name or '',
-            'version': self.version or '',
+            'version': self.version.value if self.version else '',
             'tag': self.tag,
-        })
+        }
+
+        for pname, pmatch, psub in patterns:
+            for key, item in context.items():
+                item = pmatch.sub(psub, item)
+                context[key + '|' + pname] = item
+        return pystache.render(value, context)
 
     def _get_prebuild(self):
         """Return the prebuild command."""
@@ -254,23 +255,13 @@ class Container(object):
                 return cmd
         return None
 
-    @property
-    def version(self):
-        """
-        Return the container version. Executes version detection if necessary. Raises
-        `VersionError` on detection failure.
-        """
-        if self._version:
-            return self._version
-        from .version import detect
-        return detect(self)
-
     def _get_fig_cfg(self, test):
         """Return the formatted fig configuration."""
         fig = self._running
         if not fig:
+            name = self._format('{{name|fig}}')
             fig = {
-                self.name: {
+                name: {
                     'image': '{{tag}}',
                 }
             }
@@ -302,25 +293,104 @@ class Container(object):
         except yaml.YAMLError as e:
             raise ContainerError("failed to encode config: {}", e)
 
-    def _get_fig_args(self, *args):
-        """
-        Given arguments to fig return the full list of arguments to pass to `subprocess.call`.
-        """
-        project = (self.user or '') + self.name
-        figcfg = os.path.join(self.path, 'fig.yml')
-        return ['fig', '-p', project, '-f', figcfg] + list(args)
-
-    @property
-    def tag(self):
-        """Return the full container tag."""
+    def _get_tag(self, version):
+        """Return the tag for a particular version."""
         tag = self.name
         if self.user:
             tag = '/'.join((self.user, tag))
-        version = self.version
         if version:
-            tag = ':'.join((tag, version))
+            tag = ':'.join((tag, version.value))
         return tag
 
+    def to_dict(self, raw=False):
+        """
+        Return the container as a dictionary. If `raw` is `True` then the un-evaluated
+        configuraiton is returned.
+        """
+        def rel(base, path):
+            if path and path.startswith(base):
+                path = '.' + path[len(base):]
+            return path
+
+        cwd = os.getcwd()
+        if raw:
+            data = {'name': self.name}
+            if self.user:
+                data['user'] = self.user
+            if self.verify:
+                data['verify'] = self.verify
+            if self._version:
+                data['version'] = self._version.value
+            if self._prebuild:
+                data['prebuild'] = self._prebuild
+            if self._running:
+                data['running'] = self._running
+            if self._testing:
+                data['testing'] = self._testing
+
+            path = rel(cwd, self.path)
+            if path and path != '.':
+                data['path'] = path
+        else:
+            version = self.version
+            versions = self.versions
+            data = {
+                'tag': self.tag,
+                'tags': {self._get_tag(v): v.verified for v in versions},
+                'user': self.user,
+                'name': self.name,
+                'verify': self.verify,
+                'version': self.version.value if version else None,
+                'versions': {v.value: v.verified for v in versions},
+                'running': self._get_fig_cfg(False),
+                'testing': self._get_fig_cfg(True),
+                'is-built': self.is_built(),
+                'is-running': self.is_running(),
+                'is-verified': self.is_verified(),
+            }
+            path = rel(cwd, self.path)
+            if path and path != '.':
+                data['path'] = path
+            prebuild = rel(cwd, self._get_prebuild())
+            if prebuild and prebuild != '.':
+                data['prebuild'] = prebuild
+        return data
+
+    @property
+    def version(self):
+        """
+        Return the best version for the container or `None` if unable to detect any version.
+        Executes version detection if not explicitly defined in the constructor. Raises
+        `VersionError` on detection failure.
+        """
+        if self._version:
+            return self._version
+        return get_best_version(self, self.verify)
+
+    @property
+    def versions(self):
+        """
+        Return all available versions of the container. Executes version detection if the version
+        is not explicitly defined in the constructor. Raises `VersionError` on detection
+        failure.
+        """
+        if self._version:
+            return [self._version]
+        versions = get_all_versions(self)
+        if self.verify:
+            versions = filter(lambda v: v.verified, versions)
+        return list(versions)
+
+    @property
+    def tag(self):
+        """Return the full container tag. This is tagged with the 'best' version."""
+        return self._get_tag(self.version)
+
+    @property
+    def tags(self):
+        """Return a list of all container tags."""
+        return [self._get_tag(v) for v in self.versions]
+
     @property
     def dockerfile(self):
         """Return the expected path to the Dockerfile for this container."""
@@ -332,35 +402,37 @@ class Container(object):
         existing image is up to date.
         """
         try:
-            with open(os.devnull, 'w') as devnull:
-                code = subprocess.call(['docker', 'inspect', self.tag],
-                                       cwd=self.path, stderr=devnull,
-                                       stdout=devnull, stdin=devnull)
-                return code == 0
+            built = True
+            docker_inspect = self._docker('inspect')
+            for tag in self.tags:
+                built = built and docker_inspect(tag).quiet(True) == 0
+            return built
         except OSError:
-            raise ContainerError("{} built check failed: docker is not available".format(self.tag))
+            msg = "{} build check failed: docker command is not available"
+            raise ContainerError(msg.format(self.tag))
 
     def is_running(self):
         """Return True if the container is running."""
-        figcfg = os.path.join(self.path, 'fig.yml')
-        if not os.path.isfile(figcfg):
+        if not os.path.isfile(self._fig_config):
             return False
+
         try:
-            args = self._get_fig_args('ps', '-q')
-            out = subprocess.check_output(args, cwd=self.path).strip()
-            if out:
-                return True
-            return False
-        except subprocess.CalledProcessError:
+            out, _, exitcode = self._fig('ps', '-q').capture()
+        except OSError as e:
+            msg = "{} running check failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
+
+        if exitcode != 0:
             msg = "{} running check failed: fig returned non-zero exit code"
             raise ContainerError(msg.format(self.tag))
-        except OSError:
-            raise ContainerError("{} running check failed: fig is not available".format(self.tag))
+        if out.strip():
+            return True
+        return False
 
     def is_verified(self):
         """Return True if this version of the container is verified."""
-        from .version import verify
-        return verify(self)
+        version = self.version
+        return version and version.verified
 
     def prebuild(self):
         """Execute container prebuild step."""
@@ -382,16 +454,24 @@ class Container(object):
         except subprocess.CalledProcessError:
             msg = "{} prebuild failed: command returned non-zero exit code"
             raise ContainerError(msg.format(self.tag))
-        except OSError:
-            raise ContainerError("{} prebuild failed: command is not available".format(self.tag))
+        except OSError:  # pragma: no cover
+            msg = "{} prebuild failed: command is not available"
+            raise ContainerError(msg.format(self.tag))
 
-    def build(self, prebuild=None, nocache=None):
+    def build(self, prebuild=None, nocache=None, latest=None):
         """
         Build the container. If `prebuild` is `True` (the default) the container's prebuild scripts
         are run. The Docker cache may be invalidated prior to the build by setting `nocache` to
-        `True`.
+        `True`. The image will be tagged with each version output by `versions`. It may also be
+        tagged as 'latest' by setting `latest` to `True`.
 
-        Raises `ContainerError` on failure.
+        Raises `ContainerError` on failure. Failures include:
+        - No Dockerfile.
+        - Unverified version when `verify` is `True`.
+        - Call to `docker build` fails.
+        - The `docker` command is not found.
+
+        Return a list of built tags.
         """
         if prebuild is None:
             prebuild = True
@@ -400,97 +480,131 @@ class Container(object):
 
         if not os.path.exists(self.dockerfile):
             raise ContainerError("{} build failed: no Dockerfile".format(self.tag))
+        if self.verify and not self.is_verified():
+            raise ContainerError("{} build failed: not verified".format(self.tag))
+
+        tags = self.tags
+        if latest:
+            tags.append(self._get_tag(latest_version))
+        if not tags:
+            raise ContainerError("{} build failed: no tags to build".format(self.tag))
+        image = tags.pop()
+
         if prebuild:
             self.prebuild()
 
-        args = ['docker', 'build', '-t', self.tag]
+        docker_build = self._docker('build')
         if self.allow_remove:
-            args += ['--rm', '--force-rm']
+            docker_build = docker_build('--rm', '--force-rm')
         if nocache:
-            args.append('--no-cache')
-        args.append(self.path)
+            docker_build = docker_build('--no-cache')
 
         try:
-            subprocess.check_call(args, cwd=self.path)
-        except subprocess.CalledProcessError:
-            msg = "{} build failed: docker returned non-zero exit code"
-            raise ContainerError(msg.format(self.tag))
-        except OSError:
-            raise ContainerError("{} build failed: docker is not available".format(self.tag))
+            if docker_build('-t', image, '.').call() != 0:
+                msg = "{} build failed: docker build returned non-zero exit code"
+                raise ContainerError(msg.format(self.tag))
 
-    def push(self):
-        """Push the built container to its registry."""
+            docker_tag = self._docker('tag', image)
+            for tag in tags:
+                if docker_tag(tag).call() != 0:
+                    msg = "{} build failed: docker tag returned non-zero exit code"
+                    raise ContainerError(msg.format(self.tag))
+            return tags + [image]
+        except OSError as e:
+            msg = "{} build failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
+
+    def push(self, latest=None):
+        """
+        Push the built container to its registry. Also push the 'latest' tag if `latest` is
+        True. Return a list of pushed tags.
+        """
+        if self.verify and not self.is_verified():
+            raise ContainerError("{} push failed: not verified".format(self.tag))
         if not self.is_built():
             raise ContainerError("{} push failed: image not built".format(self.tag))
-        try:
-            subprocess.check_call(['docker', 'push', self.tag], cwd=self.path)
-        except subprocess.CalledProcessError:
-            msg = "{} push failed: docker returned non-zero exit code"
-            raise ContainerError(msg.format(self.tag))
-        except OSError:
-            raise ContainerError("{} push failed: docker is not available".format(self.tag))
 
-    def remove(self):
-        """
-        Remove the container from docker. Raise a ContainerError if the remove fails.
-        """
+        tags = self.tags
+        if latest:
+            tags.append(self._get_tag(latest_version))
+        if not tags:
+            raise ContainerError("{} push failed: no tags to push".format(self.tag))
+
         try:
-            subprocess.check_call(['docker', 'rmi', self.tag], cwd=self.path)
-        except subprocess.CalledProcessError:
-            msg = "{} remove failed: docker returned non-zero exit code"
-            raise ContainerError(msg.format(self.tag))
-        except OSError:
-            raise ContainerError("{} remove failed: docker is not available".format(self.tag))
+            docker_push = self._docker('push')
+            for tag in tags:
+                if docker_push(tag).call() != 0:
+                    msg = "{} push failed: docker returned non-zero exit code"
+                    raise ContainerError(msg.format(self.tag))
+            return tags
+        except OSError as e:
+            msg = "{} push failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
+
+    def remove(self, latest=None):
+        """
+        Remove the container from docker. If `latest` is `True` also remove the container tagged as
+        'latest'. Return a list of two-tuples containing the tag name and `True` if removed. Raise
+        a `ContainerError` if the docker command is missing.
+        """
+        tags = self.tags
+        if latest:
+            tags.append(self._get_tag(latest_version))
+
+        removed = []
+        try:
+            docker_rmi = self._docker('rmi')
+            for tag in tags:
+                exitcode = docker_rmi(tag).call()
+                removed.append((tag, exitcode == 0))
+            return removed
+        except OSError as e:
+            msg = "{} remove failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
 
     def run(self, detach=None):
         """
-        Run the container. Calls `fig up` to start the container and its dependencies. As such fig
-        and a fig.yml are required to be present. After running the containers are removed using
-        `fig rm`. If `detach` is set to `True` then the containers will be run in the background.
-        They must be stopped by calling `kill` for the `fig rm` cleanup to take place.
+        Run the container. Calls `fig up` to start the container and its dependencies. After
+        running the containers are removed using `fig rm`. If `detach` is set to `True` then the
+        containers will be run in the background.  They must be stopped by calling `kill` for the
+        `fig rm` cleanup to take place.
         """
+        if self.verify and not self.is_verified():
+            raise ContainerError("{} run failed: not verified".format(self.tag))
+
         self._write_fig_cfg(False)
+        fig_up = self._fig('up')
+        if detach:
+            fig_up = fig_up('-d')
+
         try:
-            args = ['up']
-            if detach:
-                args.append('-d')
-            args = self._get_fig_args(*args)
-            subprocess.check_call(args, cwd=self.path)
-        except subprocess.CalledProcessError:
-            raise ContainerError("{} run failed: fig returned non-zero exit code".format(self.tag))
-        except OSError:
-            raise ContainerError("{} run failed: fig is not available".format(self.tag))
-        except KeyboardInterrupt:
-            pass
-        finally:
-            if not detach and self.allow_remove:
-                args = self._get_fig_args('rm', '--force', '-v')
-                try:
-                    subprocess.call(args, cwd=self.path)
-                except OSError:
-                    pass
+            try:
+                if fig_up.call() != 0:
+                    msg = "{} run failed: fig returned non-zero exit code"
+                    raise ContainerError(msg.format(self.tag))
+            finally:
+                if not detach and self.allow_remove:
+                    self._fig_remove.quiet()
+        except OSError as e:
+            msg = "{} run failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
 
     def kill(self):
         """
         Kill the running container. Calls `fig kill` to stop the container and its dependencies.
-        Once stopped a `fig rm` is issued to clean up the stopp containers.
+        Once stopped a `fig rm` is issued to clean up the stopped containers.
         """
         try:
-            if self.is_running():
-                args = self._get_fig_args('kill')
-                subprocess.check_call(args, cwd=self.path)
-        except subprocess.CalledProcessError:
-            msg = "{} kill failed: fig returned non-zero exit code"
-            raise ContainerError(msg.format(self.tag))
-        except OSError:
-            raise ContainerError("{} kill failed: fig is not available".format(self.tag))
-        finally:
-            if self.allow_remove:
-                args = self._get_fig_args('rm', '--force', '-v')
-                try:
-                    subprocess.call(args, cwd=self.path)
-                except OSError:
-                    pass
+            try:
+                if self._fig('kill').call() != 0:
+                    msg = "{} kill failed: fig returned non-zero exit code"
+                    raise ContainerError(msg.format(self.tag))
+            finally:
+                if self.allow_remove:
+                    self._fig_remove.quiet()
+        except OSError as e:
+            msg = "{} kill failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
 
     def test(self, rebuild=None):
         """
@@ -500,39 +614,39 @@ class Container(object):
         Raise `TestError` if the test has an exit code other then 0. Raise ContainerError on all
         other failures.
         """
-        if not self._testing or 'test' not in self._testing:
-            raise ContainerError("{} test failed: no test container")
+        config = self._get_fig_cfg(True)
+        if 'test' not in config:
+            msg = "{} test failed: no test container"
+            raise ContainerError(msg.format(self.tag))
 
         self._write_fig_cfg(True)
-        if rebuild and 'build' in self._testing['test']:
+        services = set(config.keys()) - set(['test'])
+
+        try:
+            # rebuild the test container if requested
+            if rebuild and 'build' in config['test']:
+                if self._fig('build', 'test').call() != 0:
+                    msg = "{} test failed: failed to rebuild test container"
+                    raise ContainerError(msg.format(self.tag))
+
             try:
-                args = self._get_fig_args('build', 'test')
-                subprocess.check_call(args, cwd=self.path)
-            except subprocess.CalledProcessError:
-                msg = "{} test failed: failed to rebuild test container"
-                raise ContainerError(msg.format(self.tag))
-            except OSError:
-                raise ContainerError("{} test failed: fig is not available".format(self.tag))
+                # bring up all available services
+                if self._fig('up', '-d', *services).call() != 0:
+                    msg = "{} test failed: failed to start dependencies"
+                    raise ContainerError(msg.format(self.tag))
 
-        try:
-            services = set(self._get_fig_cfg(True).keys()) - {'test'}
-            args = self._get_fig_args('up', '-d', *services)
-            subprocess.check_call(args, cwd=self.path)
-        except subprocess.CalledProcessError:
-            raise ContainerError("{} test failed: failed to start dependencies".format(self.tag))
-        except OSError:
-            raise ContainerError("{} test failed: fig is not available".format(self.tag))
-
-        try:
-            if self.allow_remove:
-                args = self._get_fig_args('run', '--rm', 'test')
-            else:
-                args = self._get_fig_args('run', 'test')
-            exit_code = subprocess.call(args, cwd=self.path)
-            if exit_code != 0:
-                msg = "{} test failed: returned non-zero exit code {}"
-                raise TestError(msg.format(self.tag, exit_code), exit_code)
-        except OSError:
-            raise ContainerError("{} test failed: fig is not available".format(self.tag))
-        finally:
-            self.kill()
+                # run the test container
+                fig_run = self._fig('run')
+                if self.allow_remove:
+                    fig_run = fig_run('--rm')
+                exitcode = fig_run('test').call()
+                if exitcode != 0:
+                    msg = "{} test failed: returned non-zero exit code {}"
+                    raise TestError(msg.format(self.tag, exitcode), exitcode)
+            finally:
+                self._fig('kill').call()
+                if self.allow_remove:
+                    self._fig_remove.quiet()
+        except OSError as e:
+            msg = "{} test failed: {}"
+            raise ContainerError(msg.format(self.tag, e))
